import csv
import re

from aerofiles.errors import ParserError

from .common import SeeYouFileFormat


RE_COUNTRY = re.compile(r'^([\w]{2})?$', re.I)
RE_LATITUDE = re.compile(r'^([\d]{2})([\d]{2}\.[\d]{3})([NS])$', re.I)
RE_LONGITUDE = re.compile(r'^([\d]{3})([\d]{2}\.[\d]{3})([EW])$', re.I)
RE_ELEVATION = re.compile(r'^(-?[\d]*(?:\.[\d]+)?)\s?(m|ft)?$', re.I)
RE_RUNWAY_LENGTH = re.compile(r'^(?:([\d]+(?:\.[\d]+)?)\s?(ml|nm|m)?)?$', re.I)
RE_FREQUENCY = re.compile(r'^1[\d]{2}\.[\d]+?$')
RE_DISTANCE = re.compile(r'^(-?[\d]*(?:\.[\d]+)?)\s?(m|ft|km|ml|nm)?$', re.I)


class Reader:
    """
    A reader for the SeeYou CUP waypoint file format.

    see http://download.naviter.com/docs/CUP-file-format-description.pdf
    """

    def __init__(self, fp=None):
        self.fp = fp
        self.headers = SeeYouFileFormat.HEADER_11

    def __iter__(self):
        return self.next()

    def next(self):
        waypoints = self.read(self.fp)['waypoints']
        for waypoint in waypoints:
            yield waypoint

    def read(self, fp):
        waypoints = []
        tasks = []
        task_information = False
        for fields in csv.reader(fp):
            if fields == ["-----Related Tasks-----"]:
                task_information = True
                continue
            if task_information:
                if fields[0].lower() == 'options':
                    tasks[-1]['Options'] = self.decode_task_options(fields)
                elif fields[0].lower().startswith('obszone'):
                    tasks[-1]['obs_zones'].append(self.decode_task_obs_zone(fields))
                else:
                    tasks.append(
                        {
                            'name': self.decode_task_name(fields),
                            'waypoints': self.decode_task_waypoints(fields),
                            'options': None,
                            'obs_zones': []
                        }
                    )
            else:
                waypoint = self.decode_waypoint(fields)
                if waypoint:
                    waypoints.append(waypoint)

        return dict(waypoints=waypoints, tasks=tasks)

    def decode_waypoint(self, fields):
        # Identify header line
        if all(f in fields for f in SeeYouFileFormat.HEADER_11):
            # Take headers from file instead of standard headers
            self.headers = fields
            return

        # Ignore empty lines
        num_fields = len(fields)
        if num_fields == 0:
            return

        # Ignore comments
        if fields[0].startswith('*'):
            return

        fields = [field.strip() for field in fields]

        return {
            'name': self.decode_name(fields[self.headers.index('name')]),
            'code': self.decode_code(fields[self.headers.index('code')]),
            'country': self.decode_country(fields[self.headers.index('country')]),
            'latitude': self.decode_latitude(fields[self.headers.index('lat')]),
            'longitude': self.decode_longitude(fields[self.headers.index('lon')]),
            'elevation': self.decode_elevation(fields[self.headers.index('elev')]),
            'style': self.decode_style(fields[self.headers.index('style')]),
            'runway_direction': self.decode_runway_direction(fields[self.headers.index('rwdir')]),
            'runway_length': self.decode_runway_length(fields[self.headers.index('rwlen')]),
            'runway_width': self.decode_runway_length(fields[self.headers.index('rwwidth')]) if 'rwwidth' in self.headers else None,
            'frequency': self.decode_frequency(fields[self.headers.index('freq')]),
            'description': self.decode_description(fields[self.headers.index('desc')]),
            'userdata': self.decode_code(fields[self.headers.index('userdata')]) if 'userdata' in self.headers else None,
            'pics': self.decode_pics(fields[self.headers.index('pics')]) if 'pics' in self.headers else [],
        }

    def decode_name(self, name):
        if not name:
            raise ParserError('Name field must not be empty')

        return name

    def decode_code(self, code):
        if not code:
            return None

        return code

    def decode_country(self, country):
        if RE_COUNTRY.match(country):
            return country
        else:
            raise ParserError('Invalid country code')

    def decode_latitude(self, latitude):
        match = RE_LATITUDE.match(latitude)
        if not match:
            raise ParserError('Reading latitude failed')

        latitude = int(match.group(1)) + float(match.group(2)) / 60.

        if not (0 <= latitude <= 90):
            raise ParserError('Latitude out of bounds')

        if match.group(3).upper() == 'S':
            latitude = -latitude

        return latitude

    def decode_longitude(self, longitude):
        match = RE_LONGITUDE.match(longitude)
        if not match:
            raise ParserError('Reading longitude failed')

        longitude = int(match.group(1)) + float(match.group(2)) / 60.

        if not (0 <= longitude <= 180):
            raise ParserError('Longitude out of bounds')

        if match.group(3).upper() == 'W':
            longitude = -longitude

        return longitude

    def decode_elevation(self, elevation):
        match = RE_ELEVATION.match(elevation)
        if not match:
            raise ParserError('Reading elevation failed')

        try:
            value = float(match.group(1))
        except ValueError:
            value = None

        unit = match.group(2)
        if unit and unit.lower() not in ('m', 'ft'):
            raise ParserError('Unknown elevation unit')

        return {
            'value': value,
            'unit': unit,
        }

    def decode_style(self, style):
        try:
            style = int(style)
        except ValueError:
            raise ParserError('Reading style failed')

        if not (1 <= style <= 17):
            style = 0

        return style

    def decode_runway_direction(self, runway_direction):
        if not runway_direction:
            return None

        try:
            runway_direction = int(runway_direction)
        except ValueError:
            raise ParserError('Reading runway direction failed')

        return runway_direction

    def decode_runway_length(self, runway_length):
        if not runway_length:
            return {
                'value': None,
                'unit': None,
            }

        match = RE_RUNWAY_LENGTH.match(runway_length)
        if not match:
            raise ParserError('Reading runway length/width failed')

        try:
            value = float(match.group(1))
        except ValueError:
            value = None

        unit = match.group(2)
        if unit and unit.lower() not in ('m', 'nm', 'ml'):
            raise ParserError('Unknown runway length/width unit')

        return {
            'value': value,
            'unit': unit,
        }

    def decode_frequency(self, frequency):
        if not frequency:
            return None

        if not RE_FREQUENCY.match(frequency):
            raise ParserError('Reading frequency failed')

        return frequency

    def decode_description(self, description):
        if not description:
            return None

        return description

    def decode_pics(self, pics):
        if not pics:
            return []

        return pics.split(",")

    def decode_task_options(self, fields):
        if not fields[0] == "Options":
            return

        task_options = {
            'no_start': None,
            'task_time': None,
            'wp_dis': False,
            'near_dis': None,
            'near_alt': None,
            'min_dis': False,
            'random_order': False,
            'max_pts': None,
            'before_pts': None,
            'after_pts': None,
            'bonus': None
        }

        for field in fields[1:]:
            field_type, field_entry = field.split("=")

            if field_type == 'NoStart':
                task_options['no_start'] = field_entry
            elif field_type == 'TaskTime':
                task_options['task_time'] = field_entry
            elif field_type == 'WpDis':
                task_options['wp_dis'] = field_entry == "True"
            elif field_type == 'NearDis':
                task_options['near_dis'] = self.decode_distance(field_entry)
            elif field_type == 'NearAlt':
                task_options['near_alt'] = self.decode_distance(field_entry)
            elif field_type == 'MinDis':
                task_options['min_dis'] = field_entry == "True"
            elif field_type == 'RandomOrder':
                task_options['random_order'] = field_entry == "True"
            elif field_type == 'MaxPts':
                task_options['max_pts'] = int(field_entry)
            elif field_type == 'BeforePts':
                task_options['before_pts'] = int(field_entry)
            elif field_type == 'AfterPts':
                task_options['after_pts'] = int(field_entry)
            elif field_type == 'Bonus':
                task_options['bonus'] = int(field_entry)
            else:
                raise Exception('Input contains unsupported option %s' % field)

        return task_options

    def decode_task_obs_zone(self, fields):
        task_obs_zone = {
            'obs_zone': None,
            'style': None,
            'r1': None,
            'a1': None,
            'r2': None,
            'a2': None,
            'a12': None,
            'line': False,
            'move': False,
            'reduce': False
        }

        for field in fields:
            field_type, field_entry = field.split("=")

            if field_type.lower() == 'obszone':
                task_obs_zone['obs_zone'] = int(field_entry)
            elif field_type == 'Style':
                task_obs_zone['style'] = int(field_entry)
            elif field_type == 'A1':
                task_obs_zone['a1'] = int(field_entry)
            elif field_type == 'A2':
                task_obs_zone['a2'] = int(field_entry)
            elif field_type == 'A12':
                task_obs_zone['a12'] = int(field_entry)
            elif field_type == 'R1':
                task_obs_zone['r1'] = self.decode_distance(field_entry)
            elif field_type == 'R2':
                task_obs_zone['r2'] = self.decode_distance(field_entry)
            elif field_type == 'Line' and field_entry == "1":
                task_obs_zone['line'] = True
            elif field_type == 'Move' and field_entry == "1":
                task_obs_zone['move'] = True
            elif field_type == 'Reduce' and field_entry == "1":
                task_obs_zone['reduce'] = True
            else:
                raise Exception('A taskpoint may not contain key %s' % field_type)

        return task_obs_zone

    def decode_task_name(self, fields):
        return fields[0]

    def decode_task_waypoints(self, fields):
        return fields[1::]

    def decode_distance(self, distance_str):
        if not distance_str:
            return {
                'value': None,
                'unit': None,
            }

        match = RE_DISTANCE.match(distance_str)
        if not match:
            raise ParserError('Reading neardis failed')

        try:
            value = float(match.group(1))
        except ValueError:
            value = None

        unit = match.group(2)
        if unit and unit.lower() not in ('m', 'ft', 'km', 'ml', 'nm'):
            raise ParserError('Unknown distance unit')
        return {
            'value': value,
            'unit': unit,
        }



from aerofiles.errors import ParserError
from aerofiles.util import units

from .common import WaypointStyle
from .reader import Reader


WAYPOINT_STYLE_MAPPING = {
    WaypointStyle.NORMAL: [],
    WaypointStyle.AIRFIELD_GRASS: ['landable', 'airfield'],
    WaypointStyle.OUTLANDING: ['landable'],
    WaypointStyle.GLIDERSITE: ['landable', 'airfield', 'glidersite'],
    WaypointStyle.AIRFIELD_SOLID: ['landable', 'airfield'],
    WaypointStyle.MOUNTAIN_PASS: ['mountain-pass'],
    WaypointStyle.MOUNTAIN_TOP: ['mountain-top'],
    WaypointStyle.SENDER: ['sender'],
    WaypointStyle.VOR: ['vor'],
    WaypointStyle.NDB: ['ndb'],
    WaypointStyle.COOL_TOWER: ['tower'],
    WaypointStyle.DAM: ['dam'],
    WaypointStyle.TUNNEL: ['tunnel'],
    WaypointStyle.BRIDGE: ['bridge'],
    WaypointStyle.POWER_PLANT: ['power-plant'],
    WaypointStyle.CASTLE: ['castle'],
    WaypointStyle.INTERSECTION: ['intersection'],
}

UNITS_MAPPING = {
    'm': units.METER,
    'ft': units.FEET,
    'km': units.KILOMETER,
    'ml': units.STATUTE_MILE,
    'nm': units.NAUTICAL_MILE,
}


class Converter:
    """
    A reader for the SeeYou CUP waypoint file format.

    see http://www.keepitsoaring.com/LKSC/Downloads/cup_format.pdf
    """

    def __init__(self, fp):
        self.fp = fp

    def __iter__(self):
        return self.next()

    def next(self):
        for old in Reader(self.fp):
            new = self.convert_waypoint(old)
            if new:
                yield new

    def convert_waypoint(self, old):
        waypoint = {}

        waypoint['name'] = old['name']
        waypoint['shortname'] = old['code']
        waypoint['country'] = old['country']
        waypoint['description'] = old['description']

        waypoint['latitude'] = old['latitude']
        waypoint['longitude'] = old['longitude']

        waypoint['elevation'] = self.convert_elevation(old['elevation'])

        if old['style'] not in WAYPOINT_STYLE_MAPPING:
            raise ParserError('Unknown waypoint style')

        waypoint['classifiers'] = set(WAYPOINT_STYLE_MAPPING[old['style']])

        if 'landable' in waypoint['classifiers']:
            waypoint['icao'] = None
            waypoint['runways'] = self.convert_runways(
                old['style'], old['runway_direction'], old['runway_length'])
            waypoint['frequencies'] = self.convert_frequencies(
                old['frequency'])

        return waypoint

    def convert_elevation(self, elevation):
        if elevation['value'] is None and elevation['unit'] is None:
            return None

        unit = UNITS_MAPPING.get(elevation['unit'].lower(), units.METER)
        return units.to_SI(elevation['value'] or 0, unit)

    def convert_runways(self, style, dir, length):
        runways = []
        runway = {}

        if style == WaypointStyle.AIRFIELD_GRASS:
            runway['surface'] = 'grass'
        elif style == WaypointStyle.AIRFIELD_SOLID:
            runway['surface'] = 'solid'

        if dir is not None:
            runway['directions'] = [dir % 360, (dir + 180) % 360]

        if length['value']:
            unit = UNITS_MAPPING.get(length['unit'].lower(), units.METER)
            runway['length'] = units.to_SI(length['value'], unit)

        if runway:
            runways.append(runway)

        return runways

    def convert_frequencies(self, frq):
        if not frq:
            return []

        if len(frq) < 7:
            frq += \
                '5' if frq.endswith('2') or frq.endswith('7') else '0'

        return [{
            'frequency': frq,
        }]



import re

from .reader import Reader


RE_ICAO = re.compile(r'^[\w\d]{4}$')

RE_CLASSIFIERS = [
    (re.compile(r'\bA[\d]+ B?AB[\d]*[abc]?\b'), ['highway-exit']),
    (re.compile(r'\bA[\d]+XA[\d]+\b'), ['highway-interchange']),
    (re.compile(r'\bA[\d]+YA[\d]+\b'),
        ['highway-interchange', 'y-interchange']),
    (re.compile(r'\bSTR\b'), ['road']),
    (re.compile(r'\bSX\b'), ['road-interchange']),
    (re.compile(r'\bSY\b'), ['road-interchange', 'y-interchange']),
    (re.compile(r'\bEX\b'), ['railway-interchange']),
    (re.compile(r'\bEY\b'), ['railway-interchange', 'y-interchange']),
    (re.compile(r'\bTR\b'), ['gas-station']),
    (re.compile(r'\b(BF|RS|RAIL(WAY)? STATION|GARE)\b'), ['railway-station']),
    (re.compile(r'\b(BR|BRIDGE|VIADUCT|PONT)\b'), ['bridge']),
    (re.compile(r'\b(TV|TURM|TWR|TOWER|SCHORNSTEIN)\b'), ['tower']),
    (re.compile(r'\b(CHURCH|KIRCHE|KERK|KIRKE|EGLISE|TEMPLE|TEMPLOM|DOM|'
                r'KLOSTER)\b'), ['church']),
    (re.compile(r'\b(HANGAR|HALLE)\b'), ['hangar']),
    (re.compile(r'\b(CASTLE|BURG|SCHLOSS|KOSTEL|KOSTOL|FESTUNG)\b'),
        ['castle']),
    (re.compile(r'\bVILLAGE\b'), ['village']),
    (re.compile(r'\b(TOP|PEAK|BERG|GIPFEL|MOUNTAIN|SPITZE|SUMMIT)\b'),
        ['mountain-top']),
    (re.compile(r'^(MT|MOUNT|MONT|MONTE)\b'), ['mountain']),
    (re.compile(r'\b(COL|PASS)\b'), ['mountain-pass']),
    (re.compile(r'\b(KW|KKW|POWER (PLANT|ST|STA|STN|STATION))\b'),
        ['power-plant']),
    (re.compile(r'\b(DAM|(STAU)?DAMM|STAUMAUER)\b'), ['dam']),
    (re.compile(r'\b(SILO)\b'), ['silo']),
    (re.compile(r'\b(VULCAN|VULKAN|VOLCANO)\b'), ['volcano']),
    (re.compile(r'\b(LAKE|LAC|(STAU)?SEE|TEICH)\b'), ['lake']),
    (re.compile(r'\b(SENDER)\b'), ['sender']),
    (re.compile(r'\b(VOR)\b'), ['vor']),
    (re.compile(r'\b(NDB)\b'), ['ndb']),
    (re.compile(r'\b(TUNNEL)\b'), ['tunnel']),
    (re.compile(r'\b(PFLICHTMELDEPUNKT)\b'), ['reporting-point']),
]


class Converter:
    """
    A reader for the WELT2000 waypoint file forrmat.

    see http://www.segelflug.de/vereine/welt2000/download/WELT2000-SPEC.TXT
    """

    def __init__(self, fp):
        self.fp = fp

    def __iter__(self):
        return self.next()

    def next(self):
        for old in Reader(self.fp):
            new = self.convert_waypoint(old)
            if new:
                yield new

    def convert_waypoint(self, old):
        waypoint = {}

        waypoint['name'] = old['text']
        waypoint['shortname'] = old['shortform']
        waypoint['country'] = old['country']
        waypoint['description'] = None

        waypoint['latitude'] = old['latitude']
        waypoint['longitude'] = old['longitude']

        waypoint['elevation'] = old['elevation']

        waypoint['classifiers'] = set()

        if old['is_glidersite']:
            waypoint['classifiers'].add('glidersite')

        if old['is_ulm']:
            waypoint['classifiers'].add('ulm')

        if old['is_airfield']:
            waypoint['classifiers'].add('landable')
            waypoint['classifiers'].add('airfield')

        if old['is_outlanding']:
            waypoint['classifiers'].add('landable')

        if 'landable' in waypoint['classifiers']:
            waypoint['icao'] = self.convert_icao(old['icao'])
            waypoint['runways'] = self.convert_runways(
                old['runway_surface'],
                old['runway_directions'],
                old['runway_length']
            )
            waypoint['frequencies'] = self.convert_frequencies(
                old['frequency']
            )

            if old['field_number'] is not None:
                waypoint['classifiers'].add('catalogued')
                waypoint['field_number'] = old['field_number']

        for regex, values in RE_CLASSIFIERS:
            if regex.search(old['text']):
                waypoint['classifiers'].update(values)

        return waypoint

    def convert_icao(self, icao):
        if icao and RE_ICAO.match(icao):
            return icao

    def convert_runways(self, surface, directions, length):
        runways = []
        if surface or length or directions:
            runway = {}
            if surface:
                runway['surface'] = surface
            if length:
                runway['length'] = length
            if directions:
                if len(directions) == 1:
                    runway['directions'] = [directions[0] % 360]
                else:
                    runway['directions'] = \
                        [directions[0] % 360, (directions[0] + 180) % 360]

            runways.append(runway)

        if (directions and len(directions) == 2 and
                (directions[0] + 180) % 360 != directions[1]):
            runways.append({
                'directions': [
                    directions[1] % 360,
                    (directions[1] + 180) % 360
                ],
            })

        return runways

    def convert_frequencies(self, frq):
        if not frq:
            return []

        return [{
            'frequency': frq,
        }]



# flake8: noqa

import aerofiles.flarmcfg
import aerofiles.igc
import aerofiles.openair
import aerofiles.seeyou
import aerofiles.welt2000
import aerofiles.xcsoar

from aerofiles.flarmcfg import (
    Writer as FlarmConfigWriter,
)

from aerofiles.igc import (
    Writer as IGCWriter,
)

from aerofiles.openair import (
    Reader as OpenAirReader,
)

from aerofiles.seeyou import (
    Converter as SeeYouConverter,
    Reader as SeeYouReader,
    Writer as SeeYouWriter,
)

from aerofiles.welt2000 import (
    Reader as Welt2000Reader,
    Converter as Welt2000Converter,
)

from aerofiles.xcsoar import (
    Writer as XCSoarWriter,
)




class Writer:

    """
    A writer for the Flarm configuration file format::

        with open('flarmcfg.txt', 'w') as fp:
            writer = Writer(fp)

    see `FTD-14 FLARM Configuration Specification
    <https://www.flarm.com/de/support/downloads/>`.
    """

    def __init__(self, fp=None):
        self.fp = fp

    def format_coordinate(self, value, default=None, is_latitude=True):
        if value is None:
            return default

        if is_latitude:
            if not -90 <= value <= 90:
                raise ValueError('Invalid latitude: %s' % value)

            hemisphere = 'S' if value < 0 else 'N'
            format = '%02d%05d%s'

        else:
            if not -180 <= value <= 180:
                raise ValueError('Invalid longitude: %s' % value)

            hemisphere = 'W' if value < 0 else 'E'
            format = '%03d%05d%s'

        value = abs(value)
        degrees = int(value)
        milliminutes = round((value - degrees) * 60000)
        return format % (degrees, milliminutes, hemisphere)

    def format_latitude(self, value):
        return self.format_coordinate(
            value, default='0000000N', is_latitude=True)

    def format_longitude(self, value):
        return self.format_coordinate(
            value, default='00000000E', is_latitude=False)

    def write_line(self, line):
        self.fp.write((line + u'\r\n').encode('ascii', 'replace'))

    def write_config(self, key, value):
        self.write_line('$PFLAC,S,%s,%s' % (key, value))

    def write_pilot(self, pilot):
        """
        Write the pilot name configuration::

            writer.write_pilot('Tobias Bieniek')
            # -> $PFLAC,S,PILOT,Tobias Bieniek

        :param pilot: name of the pilot
        """
        self.write_config('PILOT', pilot)

    def write_copilot(self, copilot):
        """
        Write the copilot name configuration::

            writer.write_copilot('John Doe')
            # -> $PFLAC,S,COPIL,John Doe

        :param copilot: name of the copilot
        """
        self.write_config('COPIL', copilot)

    def write_glider_type(self, glider_type):
        """
        Write the glider type configuration::

            writer.write_glider_type('Hornet')
            # -> $PFLAC,S,GLIDERTYPE,Hornet

        :param glider_type: the type of glider
        """
        self.write_config('GLIDERTYPE', glider_type)

    def write_glider_id(self, glider_id):
        """
        Write the glider registration configuration::

            writer.write_glider_id('D-4449')
            # -> $PFLAC,S,GLIDERID,D-4449

        :param glider_id: the registration of the glider
        """
        self.write_config('GLIDERID', glider_id)

    def write_competition_id(self, competition_id):
        """
        Write the competition id configuration::

            writer.write_competition_id('TH')
            # -> $PFLAC,S,COMPID,TH

        :param competition_id: competition id of the glider
        """
        self.write_config('COMPID', competition_id)

    def write_competition_class(self, competition_class):
        """
        Write the competition class configuration::

            writer.write_competition_class('Club')
            # -> $PFLAC,S,COMPCLASS,Club

        :param competition_class: competition class of the glider
        """
        self.write_config('COMPCLASS', competition_class)

    def write_logger_interval(self, interval):
        """
        Write the logger interval configuration::

            writer.write_logger_interval(4)
            # -> $PFLAC,S,LOGINT,4

        :param interval: competition class of the glider
        """
        self.write_config('LOGINT', str(interval))

    def write_task_declaration(self, description=None):
        """
        Start a new task declaration. Any old task declaration will be cleared
        by this command::

            writer.write_task_declaration('My Task')
            # -> $PFLAC,S,NEWTASK,My Task

        :param description: optional text description of task, e.g.
            "500km triangle"; can be an empty string; will be trimmed to
            50 characters
        """
        if not description:
            description = ''

        self.write_config('NEWTASK', description[0:50])

    def write_waypoint(self, latitude=None, longitude=None, description=None):
        """
        Adds a waypoint to the current task declaration. The first and the
        last waypoint added will be treated as takeoff and landing location,
        respectively.

        ::

            writer.write_waypoint(
                latitude=(51 + 7.345 / 60.),
                longitude=(6 + 24.765 / 60.),
                text='Meiersberg',
            )
            # -> $PFLAC,S,ADDWP,5107345N,00624765E,Meiersberg

        If no ``latitude`` or ``longitude`` is passed, the fields will be
        filled with zeros (i.e. unknown coordinates). This however should only
        be used for takeoff  and landing points.

        :param latitude: latitude of the point (between -90 and 90 degrees)
        :param longitude: longitude of the point (between -180 and 180 degrees)
        :param description: arbitrary text description of waypoint
        """

        if not description:
            description = ''

        latitude = self.format_latitude(latitude)
        longitude = self.format_longitude(longitude)

        self.write_config(
            'ADDWP', '%s,%s,%s' % (latitude, longitude, description[0:50])
        )

    def write_waypoints(self, points):
        """
        Write multiple task declaration points with one call::

            writer.write_waypoints([
                (None, None, 'TAKEOFF'),
                (51.40375, 6.41275, 'START'),
                (50.38210, 8.82105, 'TURN 1'),
                (50.59045, 7.03555, 'TURN 2'),
                (51.40375, 6.41275, 'FINISH'),
                (None, None, 'LANDING'),
            ])
            # -> $PFLAC,S,ADDWP,0000000N,00000000E,TAKEOFF
            # -> $PFLAC,S,ADDWP,5124225N,00624765E,START
            # -> $PFLAC,S,ADDWP,5022926N,00849263E,TURN 1
            # -> $PFLAC,S,ADDWP,5035427N,00702133E,TURN 2
            # -> $PFLAC,S,ADDWP,5124225N,00624765E,FINISH
            # -> $PFLAC,S,ADDWP,0000000N,00000000E,LANDING

        see the :meth:`~aerofiles.flarmcfg.Writer.write_waypoint` method for
        more information.

        :param points: a list of ``(latitude, longitude, text)`` tuples.
        """
        for args in points:
            if len(args) != 3:
                raise ValueError('Invalid number of task point tuple items')

            self.write_waypoint(*args)



import re

from aerofiles.errors import ParserError


RE_WHITESPACE = re.compile(r'\s+')
RE_LATITUDE = re.compile(r'^([NS])([\d]{2})([\d]{2})([\d]{2})$')
RE_LONGITUDE = re.compile(r'^([EW])([\d]{3})([\d]{2})([\d]{2})$')
RE_FIELD_NUMBER = re.compile(r'^\*FL([\d]+)$')
RE_RUNWAY_LENGTH = re.compile(r'^\s*([\d]+)\s*$')
RE_RUNWAY_DIRECTION = re.compile(r'^\s*([\d]+)\s*$')
RE_FREQUENCY = re.compile(r'^(1[\d]{2})([\d]{2})$')
RE_ELEVATION = re.compile(r'^[\s]*(-?[\d]+)')

SURFACES = {
    'A': 'asphalt',
    'C': 'concrete',
    'L': 'loam',
    'S': 'sand',
    'Y': 'clay',
    'G': 'grass',
    'V': 'gravel',
    'D': 'dirt',
    'W': 'meadow',
}


class Reader:
    """
    A reader for the WELT2000 waypoint file format.

    see https://github.com/skylines-project/welt2000
    """

    def __init__(self, fp=None):
        self.fp = fp

    def __iter__(self):
        return self.next()

    def next(self):
        for line in self.fp:
            wp = self.decode_waypoint(line)
            if wp:
                yield wp

    def decode_waypoint(self, line):
        line = line.strip()

        if not line or line.startswith('$'):
            return

        # Check valid line length
        if len(line) != 64:
            raise ParserError('Line length does not match 64')

        return {
            'shortform': self.decode_shortform(line),
            'is_airfield': self.decode_is_airfield(line),
            'is_unclear': self.decode_is_unclear(line),
            'is_outlanding': self.decode_is_outlanding(line),
            'shortform_zander': self.decode_shortform_zander(line),
            'text': self.decode_text(line),
            'icao': self.decode_icao(line),
            'is_ulm': self.decode_is_ulm(line),
            'field_number': self.decode_field_number(line),
            'is_glidersite': self.decode_is_glidersite(line),
            'runway_surface': self.decode_runway_surface(line),
            'runway_length': self.decode_runway_length(line),
            'runway_directions': self.decode_runway_directions(line),
            'frequency': self.decode_frequency(line),
            'elevation': self.decode_elevation(line),
            'elevation_proved': self.decode_elevation_proved(line),
            'latitude': self.decode_latitude(line),
            'longitude': self.decode_longitude(line),
            'ground_check_necessary': self.decode_ground_check_necessary(line),
            'better_coordinates': self.decode_better_coordinates(line),
            'country': self.decode_country(line),
            'year_code': self.decode_year_code(line),
            'source_code': self.decode_source_code(line),
        }

    def has_metadata(self, line):
        return line[23] == '#' or line[23] == '*' or line[23] == '?'

    def decode_shortform(self, line):
        return line[0:6]

    def decode_is_airfield(self, line):
        return line[5] == '1'

    def decode_is_unclear(self, line):
        return line[4] == '2'

    def decode_is_outlanding(self, line):
        return line[5] == '2'

    def decode_shortform_zander(self, line):
        if line[6] == ' ' or line[6] == '-':
            return line[7:19].rstrip()
        else:
            return line[7:19] + line[6]

    def decode_text(self, line):
        if not self.has_metadata(line):
            return line[7:41].rstrip('?! ')
        elif line[20:23] == 'GLD':
            return line[7:20].rstrip('?! ')
        else:
            return line[7:23].rstrip('?! ')

    def decode_icao(self, line):
        if (line[23] == '#' and line[24] != ' ' and
                line[27] != '?' and line[27] != '!'):
            return line[24:28].rstrip()

    def decode_is_ulm(self, line):
        return (
            line[23:27] == '*ULM' or
            line[23:27] == '#ULM' or
            line[23:28] == '# ULM'
        )

    def decode_field_number(self, line):
        match = RE_FIELD_NUMBER.match(line[23:28])
        if match:
            return int(match.group(1))

    def decode_is_glidersite(self, line):
        return (
            line[23:28] == '# GLD' or
            line[23:27] == '#GLD' or
            line[23:27] == '*GLD' or
            line[19:24] == 'GLD #' or
            line[20:24] == 'GLD#' or
            line[20:24] == 'GLD*'
        )

    def decode_runway_surface(self, line):
        if self.has_metadata(line):
            return SURFACES.get(line[28], None)

    def decode_runway_length(self, line):
        if self.has_metadata(line):
            match = RE_RUNWAY_LENGTH.match(line[29:32])
            if match:
                return int(match.group(1)) * 10

    def decode_runway_directions(self, line):
        if self.has_metadata(line):
            directions = []

            match = RE_RUNWAY_DIRECTION.match(line[32:34])
            if match:
                direction = int(match.group(1)) * 10
                directions.append(direction)

            match = RE_RUNWAY_DIRECTION.match(line[34:36])
            if match:
                direction = int(match.group(1)) * 10
                if direction not in directions:
                    directions.append(direction)

            return directions or None

    def decode_frequency(self, line):
        if self.has_metadata(line):
            match = RE_FREQUENCY.match(line[36:41])
            if match:
                frq = match.group(1) + '.' + match.group(2)
                frq += '5' if frq.endswith('2') or frq.endswith('7') else '0'
                return frq

    def decode_elevation(self, line):
        match = RE_ELEVATION.match(line[41:45])
        if match:
            return int(match.group(1))

    def decode_elevation_proved(self, line):
        return line[41] == '0'

    def decode_latitude(self, line):
        match = RE_LATITUDE.match(line[45:52])
        if not match:
            raise ParserError('Reading latitude failed')

        lat = (
            int(match.group(2)) +
            int(match.group(3)) / 60. +
            int(match.group(4)) / 3600.
        )

        if not (0 <= lat <= 90):
            raise ParserError('Latitude out of bounds')

        if match.group(1) == 'S':
            lat = -lat

        return lat

    def decode_longitude(self, line):
        match = RE_LONGITUDE.match(line[52:60])
        if not match:
            raise ParserError('Reading longitude failed')

        lon = (
            int(match.group(2)) +
            int(match.group(3)) / 60. +
            int(match.group(4)) / 3600.
        )

        if not (0 <= lon <= 180):
            raise ParserError('Longitude out of bounds')

        if match.group(1) == 'W':
            lon = -lon

        return lon

    def decode_ground_check_necessary(self, line):
        return '?' in line

    def decode_better_coordinates(self, line):
        return line[6] == '-'

    def decode_country(self, line):
        return line[60:62].strip()

    def decode_year_code(self, line):
        return line[62].strip()

    def decode_source_code(self, line):
        return line[63].strip()




class WaypointStyle:
    NORMAL = 1
    AIRFIELD_GRASS = 2
    OUTLANDING = 3
    GLIDERSITE = 4
    AIRFIELD_SOLID = 5
    MOUNTAIN_PASS = 6
    MOUNTAIN_TOP = 7
    SENDER = 8
    VOR = 9
    NDB = 10
    COOL_TOWER = 11
    DAM = 12
    TUNNEL = 13
    BRIDGE = 14
    POWER_PLANT = 15
    CASTLE = 16
    INTERSECTION = 17
    MARKER = 18
    CONTROL_POINT = 19
    PG_TAKE_OFF = 20
    PG_LANDING_ZONE = 21


class ObservationZoneStyle:
    FIXED = 0
    SYMMETRICAL = 1
    TO_NEXT_POINT = 2
    TO_PREVIOUS_POINT = 3
    TO_START_POINT = 4


class SeeYouFileFormat:
    ELEVEN = 0
    TWELVE = 1
    FORTEEN = 2

    HEADER_11 = ['name', 'code', 'country', 'lat', 'lon', 'elev', 'style', 'rwdir', 'rwlen', 'freq', 'desc']
    HEADER_12 = ['name', 'code', 'country', 'lat', 'lon', 'elev', 'style', 'rwdir', 'rwlen', 'rwwidth', 'freq', 'desc']
    HEADER_14 = ['name', 'code', 'country', 'lat', 'lon', 'elev', 'style', 'rwdir', 'rwlen', 'rwwidth', 'freq', 'desc', 'userdata', 'pics']



import sys
from pprint import pprint

from aerofiles.openair import patterns


class Reader:

    """A higher-level reader for the OpenAir airspace file format::

        with open('airspace.txt') as fp:
            reader = Reader(fp)

    see `OpenAir file format specification
    <http://www.winpilot.com/UsersGuide/UserAirspace.asp>`. It is able
    to handle standard and extended file format.

    This class should be used as a generator and will return ``(record,
    error)`` tuples for each airspace or terrain record::

        for record, error in reader:
            if error:
                raise error  # or handle it otherwise

            # handle record

    If there is a parsing error while reading a record the whole record is
    skipped and the parsing error will be returned from the generator. It is
    up to the calling code whether that parsing error should be handled as
    fatal or not.

    Airspace records have the following structure::

        {
            "type": "airspace",
            "class": "C",
            "name": "Sacramento",
            "ident": "b3836bab-6bc3-48c1-b918-01c2559e26fa",
            "ground_name": "Sacramento Radio",
            "freq": "123.456",
            "airspace_type": "CTR",
            "floor": "500ft",
            "ceiling": "UNLIM",
            "labels": [
                [39.61333, -119.76833],
            ],
            "elements": [
                ...
            ],
        }

    where "ident", "ground_name", "freq", and "airspace_type" is only
    used on extended openair file format by using "AI", "AG", "AF",
    and "AY".

    Terrain records have the following structure::

        {
            "type": "terrain",
            "open": False,
            "name": "Lake Michigan",
            "fill": [200, 200, 255],
            "outline": [0, 1, 0, 0, 255],
            "zoom": 30.0,
            "elements": [
                ...
            ],
        }

    Possible elements in both record types, where "lineno" contains
    the line number of the OpenAir file, in which the element is
    defined::

        # DP elements
        {
            "type": "point",
            "location": [39.61333, -119.76833],
            "lineno": 100,
        }

        # DA elements
        {
            "type": "arc",
            "center": [39.61333, -119.76833],
            "clockwise": True,
            "radius": 30.0,
            "start": 70.0,
            "end": 180.0,
            "lineno": 105,
        }

        # DB elements
        {
            "type": "arc",
            "center": [39.61333, -119.76833],
            "clockwise": False,
            "start": [39.61333, -119.76833],
            "end": [39.61333, -119.76833],
            "lineno": 110,
        }

        # DC elements
        {
            "type": "circle",
            "center": [39.61333, -119.76833],
            "radius": 15.0,
            "lineno": 115,
        }

        # DY elements
        {
            "type": "airway",
            "location": [39.61333, -119.76833],
            "lineno": 120,
        }

    """

    def __init__(self, fp):
        self.reader = LowLevelReader(fp)

    def __iter__(self):
        return self.next()

    def next(self):
        state = self.State()

        for line, error in self.reader:
            if error:
                yield None, error
                state.reset()
                continue

            line_type = line['type']

            if line_type in ('AC', 'AN', 'TC', 'TO') and state.is_ready():
                yield state.record, None
                state.reset()

            if not state.record:
                if line_type in ('AC', 'AN'):
                    state.reset_airspace()
                elif line_type == 'TC':
                    state.reset_terrain(False)
                elif line_type == 'TO':
                    state.reset_terrain(True)

            if state.record:
                try:
                    self.handle_line(line, state)

                except Exception as e:
                    e.lineno = self.reader.lineno
                    yield None, e
                    state.reset()

        if state.is_ready():
            yield state.record, None

    def handle_line(self, line, state):
        line_type = line['type']

        if line_type == 'AC':
            state.record["class"] = line["value"]

        elif line_type == 'AF':
            state.record["freq"] = line["value"]

        elif line_type == 'AG':
            state.record["ground_name"] = line["value"]

        elif line_type == 'AY':
            state.record["airspace_type"] = line["value"]

        elif line_type == 'AN':
            state.record["name"] = line["value"]

        elif line_type == 'AI':
            state.record["ident"] = line["value"]

        elif line_type == 'TC':
            state.record["name"] = line["value"]

        elif line_type == 'TO':
            state.record["name"] = line["value"]

        elif line_type == 'AH':
            state.record["ceiling"] = line["value"]

        elif line_type == 'AL':
            state.record["floor"] = line["value"]

        elif line_type == 'AT':
            state.record['labels'].append(line["value"])

        elif line_type == 'SP':
            state.record["outline"] = line["value"]

        elif line_type == 'SB':
            state.record["fill"] = line["value"]

        elif line_type == 'V':
            if line['name'] == 'X':
                state.center = line["value"]
            elif line['name'] == 'D':
                state.clockwise = line["value"]
            elif line['name'] == 'Z':
                state.record['zoom'] = line["value"]

        elif line_type == 'DP':
            state.add_element({
                "type": "point",
                "location": line["value"],
                "lineno": line["lineno"],
            })

        elif line_type == 'DA':
            if not state.center:
                raise ValueError('center undefined')

            state.add_element({
                "type": "arc",
                "center": state.center,
                "clockwise": state.clockwise,
                "radius": line["radius"],
                "start": line["start"],
                "end": line["end"],
                "lineno": line["lineno"],
            })

        elif line_type == 'DB':
            state.add_element({
                "type": "arc",
                "center": state.center,
                "clockwise": state.clockwise,
                "start": line["start"],
                "end": line["end"],
                "lineno": line["lineno"],
            })

        elif line_type == 'DC':
            if not state.center:
                raise ValueError('center undefined')

            state.add_element({
                "type": "circle",
                "center": state.center,
                "radius": line["value"],
                "lineno": line["lineno"],
            })

        elif line_type == 'DY':
            state.add_element({
                "type": "airway",
                "location": line["value"],
                "lineno": line["lineno"],
            })

    class State:
        def __init__(self):
            self.center = None
            self.reset()

        def is_ready(self):
            if not self.record:
                return False

            record_type = self.record.get("type")

            return record_type == "terrain" or (record_type == "airspace" and
                self.record.get("name") and self.record.get("class"))  # noqa

        def reset_airspace(self):
            self.record = {
                "type": "airspace",
                "class": None,
                "name": None,
                "floor": None,
                "ceiling": None,
                "ident": None,
                "ground_name": None,
                "freq": None,
                "airspace_type": None,
                "labels": [],
                "elements": []
            }

        def reset_terrain(self, open):
            self.record = {
                "type": "terrain",
                "open": open,
                "name": None,
                "fill": None,
                "outline": None,
                "zoom": None,
                "elements": []
            }

        def reset(self):
            self.clockwise = True
            self.record = None

        def add_element(self, element):
            self.record['elements'].append(element)


class LowLevelReader:

    """
    A low-level reader for the OpenAir airspace file format::

        with open('airspace.txt') as fp:
            reader = LowLevelReader(fp)

    see `OpenAir file format specification
    <http://www.winpilot.com/UsersGuide/UserAirspace.asp>`_

    Instances of this class read OpenAir files line by line and extract the
    information in each line as a dictionary. A line like ``AN Sacramento`` for
    example is converted to ``{"type": "AN", "value": "Sacramento"}``.

    The reader should be used as a generator and will return a ``(result,
    error)`` tuple for each line that is not empty or a comment::

        for result, error in reader:
            if error:
                raise error  # or handle it otherwise

            # handle result

    Most lines are just parsed into ``type`` and ``value`` strings. In
    addition they have ``lineno`` which contains the line number of
    the parsed OpenAir file. The following examples should show the
    lines that are parsed differently::

        # AT 39:36.8 N 119:46.1W
        {"type": "AT", "value": [39.61333, -119.76833]}

        # DA 10,320,200
        {"type": "DA", "radius": 10, "start": 320, "end": 200}

        # DC 1.5
        {"type": "DC", "value": 1.5}

        # DP 39:35:00 N 118:59:20 W
        {"type": "DP", "value": [39.58333, -118.98888]}

        # SB 200,200,255
        {"type": "SB", "value": [200, 200, 255]}

        # SP 0,1,0,0,255
        {"type": "SP", "value": [0, 1, 0, 0, 255]}

        # V D=-
        {"type": "V", "name": "D", "value": False}

        # V X=39:29.7 N 119:46.5 W
        {"type": "V", "name": "X", "value": [39.495, -119.775]}

        # V Z=100
        {"type": "V", "name": "Z", "value": 100}
    """

    def __init__(self, fp):
        self.fp = fp
        self.lineno = 0

    def __iter__(self):
        return self.next()

    def next(self):
        for line in self.fp:
            self.lineno += 1

            try:
                result = self.parse_line(line)
                if result:
                    yield (result, None)

            except Exception as e:
                e.lineno = self.lineno
                yield (None, e)

    def parse_line(self, line):
        # Ignore comments
        pos = line.find('*')
        if pos >= 0:
            line = line[:pos]

        # Strip whitespace
        line = line.strip()

        # Ignore empty lines
        if line == '':
            return None

        # Split record type from line
        record = line.split(' ', 1)

        # Find handler method
        handler = self.get_handler_method(record[0])

        # Get value from record
        value = None if len(record) < 2 else record[1]

        # Run handler method and return result
        result = handler(value)
        result["lineno"] = self.lineno

        return result

    def get_handler_method(self, type):
        handler = getattr(self, 'handle_%s_record' % type, None)
        if not handler:
            raise ValueError('unknown record type %s in line %d' % (type, self.lineno))

        return handler

    def handle_AC_record(self, value):
        return {'type': 'AC', 'value': value}

    def handle_AN_record(self, value):
        return {'type': 'AN', 'value': value}

    def handle_AF_record(self, value):
        return {'type': 'AF', 'value': value}

    def handle_AG_record(self, value):
        return {'type': 'AG', 'value': value}

    def handle_AY_record(self, value):
        return {'type': 'AY', 'value': value}

    def handle_AH_record(self, value):
        return {'type': 'AH', 'value': value}

    def handle_AI_record(self, value):
        return {'type': 'AI', 'value': value}

    def handle_AL_record(self, value):
        return {'type': 'AL', 'value': value}

    def handle_AT_record(self, value):
        return {'type': 'AT', 'value': coordinate(value)}

    def handle_TO_record(self, value):
        return {'type': 'TO', 'value': value}

    def handle_TC_record(self, value):
        return {'type': 'TC', 'value': value}

    def handle_SB_record(self, value):
        value = split(value, ',', 3, int, int, int)
        return {'type': 'SB', 'value': value}

    def handle_SP_record(self, value):
        value = split(value, ',', 5, int, int, int, int, int)
        return {'type': 'SP', 'value': value}

    def handle_V_record(self, value):
        name, value = split(value, '=', 2, str, str)

        if name == 'X':
            value = coordinate(value)

        elif name == 'Z':
            value = float(value)

        elif name == 'D':
            if value.startswith('+'):
                value = True
            elif value.startswith('-'):
                value = False
            else:
                raise ValueError('invalid direction value: %s' % value)

        return {'type': 'V', 'name': name, 'value': value}

    def handle_DP_record(self, value):
        return {'type': 'DP', 'value': coordinate(value)}

    def handle_DA_record(self, value):
        radius, start, end = split(value, ',', 3, float, float, float)

        return {
            'type': 'DA',
            'radius': radius, 'start': start, 'end': end,
        }

    def handle_DB_record(self, value):
        start, end = split(value, ',', 2, coordinate, coordinate)

        return {
            'type': 'DB',
            'start': start, 'end': end
        }

    def handle_DC_record(self, value):
        return {'type': 'DC', 'value': float(value)}

    def handle_DY_record(self, value):
        return {'type': 'DY', 'value': coordinate(value)}


def split(value, separator, num, *types):
    values = value.split(separator)
    if len(values) != num:
        raise ValueError()

    return [cast(v.strip()) for v, cast in zip(values, types)]


def coordinate(value):
    # 39:35:00 N 118:59:20 W
    match = patterns.LOCATION_FORMAT_1.match(value)
    if match:
        lat = \
            int(match.group(1)) + \
            int(match.group(2)) / 60. + \
            float(match.group(3)) / 3600.

        if match.group(4) == 'S':
            lat = -lat

        lon = \
            int(match.group(5)) + \
            int(match.group(6)) / 60. + \
            float(match.group(7)) / 3600.

        if match.group(8) == 'W':
            lon = -lon

        return [lat, lon]

    # 39:29.9 N 119:46.1 W
    match = patterns.LOCATION_FORMAT_2.match(value)
    if match:
        lat = \
            int(match.group(1)) + \
            float(match.group(2)) / 60.

        if match.group(3) == 'S':
            lat = -lat

        lon = \
            int(match.group(4)) + \
            float(match.group(5)) / 60.

        if match.group(6) == 'W':
            lon = -lon

        return [lat, lon]

    raise ValueError('invalid coordinate format: "%s"' % value)


def main(files):
    # Read every file and print result
    for file in files:
        with open(file, 'r') as fp:
            reader = Reader(fp)
            for record, error in reader:
                if error:
                    print("Error in file", file, error)
                else:
                    pprint(record)


if __name__ == "__main__":
    files = sys.argv[1:]
    if not files:
        print("Please give at least 1 openair file to read.")
        sys.exit(1)

    main(files)



import datetime

from aerofiles.util.timezone import TimeZoneFix


class Reader:
    """
    A reader for the IGC flight log file format.

    skip_duplicates flag removes trailing duplicate time entries

    Example:

    .. sourcecode:: python

        >>> with open('track.igc', 'r') as f:
        ...     parsed = Reader(skip_duplicates=True).read(f)

    """

    def __init__(self, skip_duplicates=False):
        self.reader = None
        self.skip_duplicates = skip_duplicates

    def read(self, file_obj):
        """
        Read the specified file object and return a dictionary with the parsed data.

        :param file_obj: a Python file object

        """
        self.reader = LowLevelReader(file_obj)

        logger_id = [[], None]
        fix_records = [[], []]
        task = [[], {"waypoints": []}]
        dgps_records = [[], []]
        event_records = [[], []]
        satellite_records = [[], []]
        security_records = [[], []]
        header = [[], {}]
        fix_record_extensions = [[], []]
        k_record_extensions = [[], []]
        k_records = [[], []]
        comment_records = [[], []]

        for record_type, line, error in self.reader:

            if record_type == 'A':
                if error:
                    logger_id[0].append(error)
                else:
                    logger_id[1] = line
            elif record_type == 'B':
                if error:
                    if MissingRecordsError not in fix_records[0]:
                        fix_records[0].append(MissingRecordsError)
                else:

                    # add MissingExtensionsError when fix_record_extensions contains error
                    if len(fix_record_extensions[0]) > 0 and MissingExtensionsError not in fix_records[0]:
                        fix_records[0].append(MissingExtensionsError)

                    fix_record = LowLevelReader.process_B_record(line, fix_record_extensions[1])

                    # To create "datetime" we need a date. Take it from header or previous fix:
                    if len(fix_records[1]) == 0:
                        date = header[1]["utc_date"]
                    else:
                        previous_fix = fix_records[1][-1]
                        date = previous_fix["datetime"].date()
                        time = previous_fix["datetime"].time()
                        # If time of next fix is _before_ last fix, we are now on next day
                        if fix_record["time"] < time:
                            date = date + datetime.timedelta(days=1)
                        if fix_record["time"] == time and self.skip_duplicates:
                            continue

                    fix_record["datetime"] = datetime.datetime.combine(date, fix_record["time"]).replace(tzinfo=TimeZoneFix(0))
                    if "time_zone_offset" in header[1]:
                        timezone = TimeZoneFix(header[1]["time_zone_offset"])
                        fix_record["datetime_local"] = fix_record["datetime"].astimezone(timezone)

                    fix_records[1].append(fix_record)
            elif record_type == 'C':
                task_item = line

                if error:
                    if MissingRecordsError not in task[0]:
                        task[0].append(MissingRecordsError)
                else:
                    if task_item['subtype'] == 'task_info':
                        del task_item['subtype']
                        task[1]['waypoints'] = []
                        task[1].update(task_item)
                    elif task_item['subtype'] == 'waypoint_info':
                        del task_item['subtype']
                        task[1]['waypoints'].append(task_item)

            elif record_type == 'D':
                if error:
                    if MissingRecordsError not in dgps_records[0]:
                        dgps_records[0].append(MissingRecordsError)
                else:
                    dgps_records[1].append(line)
            elif record_type == 'E':
                if error:
                    if MissingRecordsError not in event_records[0]:
                        event_records[0].append(MissingRecordsError)
                else:
                    event_records[1].append(line)
            elif record_type == 'F':
                if error:
                    if MissingRecordsError not in satellite_records[0]:
                        satellite_records[0].append(MissingRecordsError)
                else:
                    satellite_records[1].append(line)
            elif record_type == 'G':
                if error:
                    if MissingRecordsError not in security_records[0]:
                        security_records[0].append(MissingRecordsError)
                else:
                    security_records[1].append(line)
            elif record_type == 'H':
                header_item = line

                if error:
                    if MissingRecordsError not in header[0]:
                        header[0].append(MissingRecordsError(error))
                else:
                    del header_item['source']
                    header[1].update(header_item)
            elif record_type == 'I':
                if error:
                    fix_record_extensions[0].append(error)
                else:
                    fix_record_extensions[1] = line
            elif record_type == 'J':
                if error:
                    k_record_extensions[0].append(error)
                else:
                    k_record_extensions[1] = line
            elif record_type == 'K':
                if error:
                    if MissingRecordsError not in k_records[0]:
                        k_records[0].append(MissingRecordsError)
                else:

                    # add MissingExtensionsError when fix_record_extensions contains error
                    if len(k_record_extensions[0]) > 0 and MissingExtensionsError not in k_records[0]:
                        k_records[0].append(MissingExtensionsError)

                    k_record = LowLevelReader.process_K_record(line, k_record_extensions[1])
                    k_records[1].append(k_record)
            elif record_type == 'L':
                if error:
                    if MissingRecordsError not in comment_records[0]:
                        comment_records[0].append(MissingRecordsError)
                else:
                    comment_records[1].append(line)

        return dict(logger_id=logger_id,                            # A record
                    fix_records=fix_records,                        # B records
                    task=task,                                      # C records
                    dgps_records=dgps_records,                      # D records
                    event_records=event_records,                    # E records
                    satellite_records=satellite_records,            # F records
                    security_records=security_records,              # G records
                    header=header,                                  # H records
                    fix_record_extensions=fix_record_extensions,    # I records
                    k_record_extensions=k_record_extensions,        # J records
                    k_records=k_records,                            # K records
                    comment_records=comment_records,                # L records
                    )


class LowLevelReader:
    """
    A low level reader for the IGC flight log file format.

    see http://carrier.csi.cam.ac.uk/forsterlewis/soaring/igc_file_format/igc_format_2008.html
    """

    def __init__(self, file_obj):
        self.file_obj = file_obj
        self.line_number = 0

    def __iter__(self):
        return self.next()

    def next(self):
        for line in self.file_obj:
            self.line_number += 1

            record_type = line[0]

            try:
                result = self.parse_line(record_type, line) if line.strip() else None  # skip empty lines
                if result:
                    yield (record_type, result, None)
            except Exception as e:
                e.line_number = self.line_number
                yield (record_type, None, e)

    def parse_line(self, record_type, line):
        decoder = self.get_decoder_method(record_type)
        return decoder(line)

    def get_decoder_method(self, record_type):
        decoder = getattr(self, 'decode_%s_record' % record_type)
        if not decoder:
            raise ValueError('Unknown record type')

        return decoder

    @staticmethod
    def decode_A_record(line):
        id_addition = None if len(line) == 7 else line[7:].strip()
        return {
            'manufacturer': line[1:4],
            'id': line[4:7],
            'id_addition': id_addition
        }

    @staticmethod
    def decode_B_record(line):
        return {
            'time': LowLevelReader.decode_time(line[1:7]),
            'lat': LowLevelReader.decode_latitude(line[7:15]),
            'lon': LowLevelReader.decode_longitude(line[15:24]),
            'validity': line[24],
            'pressure_alt': int(line[25:30]),
            'gps_alt': int(line[30:35]),
            'start_index_extensions': 35,
            'extensions_string': line[35::].strip()
        }

    @staticmethod
    def process_B_record(decoded_b_record, fix_record_extensions):

        i = decoded_b_record['start_index_extensions']
        ext = decoded_b_record['extensions_string']

        b_record = decoded_b_record
        del b_record['start_index_extensions']
        del b_record['extensions_string']

        for extension in fix_record_extensions:
            start_byte, end_byte = extension['bytes']
            start_byte = start_byte - i - 1
            end_byte = end_byte - i - 1

            b_record.update(
                {extension['extension_type']: int(ext[start_byte:end_byte + 1])}
            )

        return b_record

    @staticmethod
    def decode_C_record(line):

        task_info = line[8].isdigit()

        if task_info:
            return {
                'subtype': 'task_info',
                'declaration_date': LowLevelReader.decode_date(line[1:7]),
                'declaration_time': LowLevelReader.decode_time(line[7:13]),
                'flight_date': LowLevelReader.decode_date(line[13:19]),
                'number': line[19:23],
                'num_turnpoints': int(line[23:25]),
                'description': line[25::].strip()
            }
        else:
            return {
                'subtype': 'waypoint_info',
                'latitude': LowLevelReader.decode_latitude(line[1:9]),
                'longitude': LowLevelReader.decode_longitude(line[9:18]),
                'description': line[18::].strip()
            }

    @staticmethod
    def decode_D_record(line):

        qualifier = line[1]
        if qualifier == '1':
            qualifier = 'GPS'
        elif qualifier == '2':
            qualifier = 'DGPS'
        else:
            raise ValueError('This qualifier is not possible')

        return {
            'qualifier': qualifier,
            'station_id': line[2:6]
        }

    @staticmethod
    def decode_E_record(line):
        return {
            'time': LowLevelReader.decode_time(line[1:7]),
            'tlc': line[7:10],
            'extension_string': line[10::].strip()

        }

    @staticmethod
    def decode_F_record(line):

        time_str = line[1:7]
        time = LowLevelReader.decode_time(time_str)

        # each satellite ID should have two digits
        if (len(line.strip()) - 7) % 2 != 0:
            raise ValueError('F record formatting is incorrect')

        satellites = []
        no_satellites = int((len(line.strip()) - 7) / 2)

        starting_byte = 7
        for satellite_index in range(no_satellites):
            satellites.append(line[starting_byte:starting_byte + 2])
            starting_byte += 2

        return {
            'time': time,
            'satellites': satellites
        }

    @staticmethod
    def decode_G_record(line):
        return line.strip()[1::]

    @staticmethod
    def decode_H_record(line):

        source = line[1]

        # three letter code
        tlc = line[2:5]

        colon = line.find(":", 5)
        if colon >= 0:
            # This is the long format
            long_name = line[5:colon].strip()
            if long_name == "":
                long_name = None
            line_value = line[colon + 1:].strip()
        else:
            long_name = None
            line_value = line[5:].strip()

        if tlc == 'DTE':
            value = LowLevelReader.decode_H_utc_date(line_value)
        elif tlc == 'FXA':
            value = LowLevelReader.decode_H_fix_accuracy(line_value)
        elif tlc == 'PLT':
            value = LowLevelReader.decode_H_pilot(line_value)
        elif tlc == 'CM2':
            value = LowLevelReader.decode_H_copilot(line_value)
        elif tlc == 'GTY':
            value = LowLevelReader.decode_H_glider_model(line_value)
        elif tlc == 'GID':
            value = LowLevelReader.decode_H_glider_registration(line_value)
        elif tlc == 'DTM':
            value = LowLevelReader.decode_H_gps_datum(line_value)
        elif tlc == 'RFW':
            value = LowLevelReader.decode_H_firmware_revision(line_value)
        elif tlc == 'RHW':
            value = LowLevelReader.decode_H_hardware_revision(line_value)
        elif tlc == 'FTY':
            value = LowLevelReader.decode_H_manufacturer_model(line_value)
        elif tlc == 'GPS':
            value = LowLevelReader.decode_H_gps_receiver(line_value)
        elif tlc == 'PRS':
            value = LowLevelReader.decode_H_pressure_sensor(line_value)
        elif tlc == 'CID':
            value = LowLevelReader.decode_H_competition_id(line_value)
        elif tlc == 'CCL':
            value = LowLevelReader.decode_H_competition_class(line_value)
        elif tlc == 'TZN':
            value = LowLevelReader.decode_H_time_zone_offset(line_value)
        elif tlc == 'MOP':
            value = LowLevelReader.decode_H_mop_sensor(line_value)
        elif tlc == 'SIT':
            value = LowLevelReader.decode_H_site(line_value)
        elif tlc == 'TZO':
            value = LowLevelReader.decode_H_time_zone_offset(line_value)
        elif tlc == 'UNT':
            value = LowLevelReader.decode_H_units_of_measure(line_value)
        elif tlc == 'FRS':
            value = LowLevelReader.decode_H_security(line_value)
        elif tlc == 'ALG':
            value = LowLevelReader.decode_H_gnss_alt(line_value)
        elif tlc == 'ALP':
            value = LowLevelReader.decode_H_pressure_alt(line_value)
        else:
            raise ValueError('Invalid h-record "%s"' % tlc)

        value.update({'source': source})

        return value

    @staticmethod
    def decode_H_utc_date(value):
        date_str = value[:6]
        return {'utc_date': LowLevelReader.decode_date(date_str)}

    @staticmethod
    def decode_H_fix_accuracy(value):
        fix_accuracy = value
        return {'fix_accuracy': None} if fix_accuracy == '' else {'fix_accuracy': int(fix_accuracy)}

    @staticmethod
    def decode_H_pilot(value):
        pilot = value
        return {'pilot': None} if pilot == '' else {'pilot': pilot}

    @staticmethod
    def decode_H_copilot(value):
        second_pilot = value
        return {'copilot': None} if second_pilot == '' else {'copilot': second_pilot}

    @staticmethod
    def decode_H_glider_model(value):
        glider_model = value
        return {'glider_model': None} if glider_model == '' else {'glider_model': glider_model}

    @staticmethod
    def decode_H_glider_registration(value):
        glider_registration = value
        if glider_registration == '':
            return {'glider_registration': None}
        else:
            return {'glider_registration': glider_registration}

    @staticmethod
    def decode_H_gps_datum(value):
        gps_datum = value
        return {'gps_datum': None} if gps_datum == '' else {'gps_datum': gps_datum}

    @staticmethod
    def decode_H_firmware_revision(value):
        firmware_revision = value
        return {'firmware_revision': None} if firmware_revision == '' else {'firmware_revision': firmware_revision}

    @staticmethod
    def decode_H_security(security):
        return {'security': None} if security == '' else {'security': security}

    @staticmethod
    def decode_H_hardware_revision(value):
        hardware_revision = value
        return {'hardware_revision': None} if hardware_revision == '' else {'hardware_revision': hardware_revision}

    @staticmethod
    def decode_H_manufacturer_model(value):
        manufacturer = None
        model = None
        manufacturer_model = value.split(',')
        if manufacturer_model[0] != '':
            manufacturer = manufacturer_model[0].strip()
        if len(manufacturer_model) == 2 and manufacturer_model[1].lstrip() != '':
            model = manufacturer_model[1].strip()
        return {'logger_manufacturer': manufacturer,
                'logger_model': model}

    @staticmethod
    def decode_H_gps_receiver(value):
        gps_sensor = value.split(',')
        manufacturer = None
        model = None
        channels = None
        max_alt = None
        for detail_index, detail in enumerate(reversed(gps_sensor)):
            if len(gps_sensor) == 1 or (len(gps_sensor) == 2 and gps_sensor[1].strip() == ''):
                manufacturer = detail.strip()
            elif len(gps_sensor) == 3:
                if detail_index == 0:
                    max_alt = detail.strip()
                elif detail_index == 1:
                    channels = detail.strip()
                else:  # detail_index == 2
                    manufacturer = detail.strip()
            elif len(gps_sensor) == 4:
                if detail_index == 0:
                    max_alt = detail.strip()
                elif detail_index == 1:
                    channels = detail.strip()
                elif detail_index == 2:
                    model = detail.strip()
                else:  # detail_index == 3
                    manufacturer = detail.strip()
            else:
                raise NotImplementedError

        # stripping of ch from '12ch'
        if channels is not None:
            # Special case for SkyBean SkyDrop/Strato vario wrongly
            # specifying channels as "cm"
            if manufacturer == 'Quectel' and model == 'L80' and channels == '22cm':
                channels = 22
            elif manufacturer == 'u-blox' and model == 'NEO-M8Q' and channels == '22cm':
                channels = 72
            else:
                if channels.endswith('ch') or channels.endswith('Ch') or channels.endswith('CH'):
                    channels = int(channels[:-2])
                else:
                    channels = int(channels)

        # stripping of max from 'max10000m'
        if max_alt is not None and max_alt.startswith('max'):
            max_alt = max_alt[3::]

        # separate unit from value
        if max_alt is not None and max_alt.endswith('m'):
            max_alt_value = int(max_alt[:-1])
            max_alt_unit = 'm'
        elif max_alt is not None and max_alt.endswith('ft'):
            max_alt_value = int(max_alt[:-2])
            max_alt_unit = 'ft'
        elif max_alt is not None:
            max_alt_value = int(max_alt)
            max_alt_unit = 'm'
        else:
            max_alt_value = None
            max_alt_unit = None

        return {
            'gps_manufacturer': manufacturer,
            'gps_model': model,
            'gps_channels': channels,
            'gps_max_alt': {
                'value': max_alt_value,
                'unit': max_alt_unit
            }
        }

    @staticmethod
    def decode_H_pressure_sensor(value):

        manufacturer = None
        model = None
        max_alt = None

        # some IGC files use colon, others don't
        pressure_sensor = value.split(',')

        if len(pressure_sensor) == 1:
            manufacturer = pressure_sensor[0].strip() if pressure_sensor[0] != '' else None
        elif len(pressure_sensor) == 2:
            manufacturer_model = pressure_sensor[0].strip().split(' ') if pressure_sensor[0] != '' else None

            if len(manufacturer_model) == 2:
                manufacturer = manufacturer_model[0]
                model = manufacturer_model[1]
            else:
                manufacturer = manufacturer_model[0]

            max_alt = pressure_sensor[1].strip() if pressure_sensor[1] != '' else None
        elif len(pressure_sensor) == 3:
            manufacturer = pressure_sensor[0].strip() if pressure_sensor[0] != '' else None
            model = pressure_sensor[1].strip() if pressure_sensor[1] != '' else None
            max_alt = pressure_sensor[2].strip() if pressure_sensor[2] != '' else None

        # stripping of max from 'max10000m'
        if max_alt is not None and max_alt.startswith('max'):
            max_alt = max_alt[3::]

        # separate unit from value
        if max_alt is not None and max_alt.endswith('m'):
            max_alt_value = int(max_alt[:-1])
            max_alt_unit = 'm'
        elif max_alt is not None and max_alt.endswith('ft'):
            max_alt_value = int(max_alt[:-2])
            max_alt_unit = 'ft'
        elif max_alt is not None:
            max_alt_value = int(max_alt)
            max_alt_unit = 'm'
        else:
            max_alt_value = None
            max_alt_unit = None

        return {
            'pressure_sensor_manufacturer': manufacturer,
            'pressure_sensor_model': model,
            'pressure_sensor_max_alt': {
                'value': max_alt_value,
                'unit': max_alt_unit
            }
        }

    @staticmethod
    def decode_H_competition_id(value):
        competition_id = value
        return {'competition_id': None} if competition_id == '' else {'competition_id': competition_id}

    @staticmethod
    def decode_H_competition_class(value):
        competition_class = value
        return {'competition_class': None} if competition_class == '' else {'competition_class': competition_class}

    @staticmethod
    def decode_H_time_zone_offset(value):
        return {'time_zone_offset': float(value)}

    @staticmethod
    def decode_H_mop_sensor(value):
        return {'mop_sensor': value}

    @staticmethod
    def decode_H_site(value):
        return {'site': value}

    @staticmethod
    def decode_H_units_of_measure(value):
        return {'units_of_measure': value.split(',')}

    @staticmethod
    def decode_H_gnss_alt(value):
        if value not in ["ELL", "GEO", "NKM", "NIL"]:
            raise ValueError('Invalid HFALG value "%s"' % value)
        return {'gnss_altitude': value}

    @staticmethod
    def decode_H_pressure_alt(value):
        if value not in ["ISA", "MSL", "NKM", "NIL"]:
            raise ValueError('Invalid HFALP value "%s"' % value)
        return {'pressure_altitude': value}

    @staticmethod
    def decode_I_record(line):
        return LowLevelReader.decode_extension_record(line)

    @staticmethod
    def decode_J_record(line):
        return LowLevelReader.decode_extension_record(line)

    @staticmethod
    def decode_K_record(line):
        return {
            'time': LowLevelReader.decode_time(line[1:7]),
            'value_string': line.strip()[7::],
            'start_index': 7
        }

    @staticmethod
    def process_K_record(decoded_k_record, k_record_extensions):

        i = decoded_k_record['start_index']
        t = decoded_k_record['time']
        val = decoded_k_record['value_string']

        k_record = {'time': t}
        for extension in k_record_extensions:

            start_byte, end_byte = extension['bytes']
            start_byte -= i
            end_byte -= i

            k_record[extension['extension_type']] = val[start_byte:end_byte]

        return k_record

    @staticmethod
    def decode_L_record(line):
        return {
            'source': line[1:4],
            'comment': line[4::].strip()
        }

    @staticmethod
    def decode_date(date_str):

        if len(date_str) != 6:
            raise ValueError('Date string does not have correct length')
        elif date_str == '000000':
            return None

        return datetime.datetime.strptime(date_str, "%d%m%y").date()

    @staticmethod
    def decode_time(time_str):

        if len(time_str) != 6:
            raise ValueError('Time string does not have correct size')

        return datetime.datetime.strptime(time_str, "%H%M%S").time()

    @staticmethod
    def decode_extension_record(line):

        no_extensions = int(line[1:3])

        if no_extensions * 7 + 3 != len(line.strip()):
            raise ValueError('I record contains incorrect number of digits')

        extensions = []
        for extension_index in range(no_extensions):
            extension_str = line[extension_index * 7 + 3:(extension_index + 1) * 7 + 3]
            start_byte = int(extension_str[0:2])
            end_byte = int(extension_str[2:4])
            tlc = extension_str[4:7]

            extensions.append({'bytes': (start_byte, end_byte), 'extension_type': tlc})

        return extensions

    @staticmethod
    def decode_latitude(lat_string):

        d = int(lat_string[0:2])
        m = float(lat_string[2:7]) / 1000
        ordinal = lat_string[7]

        latitude = d + m / 60.

        if not (0. <= latitude <= 90):
            raise ValueError('Latitude format is invalid')

        if ordinal == 'S':
            latitude = -latitude

        return latitude

    @staticmethod
    def decode_longitude(lon_string):

        d = float(lon_string[0:3])
        m = float(lon_string[3:8]) / 1000
        ordinal = lon_string[8]

        longitude = d + m / 60.

        if not (0. <= longitude <= 180):
            raise ValueError('Longitude format is invalid')

        if ordinal == 'W':
            longitude = -longitude

        return longitude


class MissingRecordsError(Exception):
    pass


class MissingExtensionsError(Exception):
    pass
